# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from bzrlib import diff, revision as _mod_revision
from bzrlib.merge import Merge3Merger


def apply_merges(old_tree, branch, merges):
    repository = branch.repository
    for revisions in merges:
        base, other = repository.revision_trees(revisions)
        merger = Merge3Merger(old_tree, old_tree, base, other,
                              this_branch=branch, do_merge=False)
        transform = merger.make_preview_transform()
        old_tree = transform.get_preview_tree()
    return old_tree


def mainline_diff(branch, old_revision, new_revision, to_file):
    other_merges = merges(branch.repository, old_revision, new_revision)
    diff_with_merges(branch, old_revision, new_revision, other_merges, to_file)


def diff_with_merges(branch, old_revision, new_revision, merges, to_file):
    repository = branch.repository
    old_tree = repository.revision_tree(old_revision)
    old_tree = apply_merges(old_tree, branch, merges)
    new_tree = repository.revision_tree(new_revision)
    diff.show_diff_trees(old_tree, new_tree, to_file, old_label='',
                         new_label='')


def ignore_heads(repository, ignore_branches, old_revision, new_revision):
    """List a bunch of revisions to ignore in a diff.

    The head revisions merged from ignore_branches into new_revision, but not
    merged into old_revision will be listed.
    """
    lcas = []
    for branch in ignore_branches:
        graph = repository.get_graph(branch.repository)
        lcas.extend(graph.find_lca(new_revision, branch.last_revision()))
    if len(lcas) == 0:
        return []
    heads = graph.heads([old_revision] + lcas)
    return [h for h in heads if h != old_revision]


def diff_ignore_branches(branch, ignore_branches, old_revision, new_revision,
                         to_file):
    """Perform a diff ignoring mergest from the specified branches."""
    repository = branch.repository
    merge_heads = ignore_heads(repository, ignore_branches, old_revision,
                               new_revision)
    graph = repository.get_graph()
    merges = iter_lca_revision(graph, old_revision, merge_heads)
    diff_with_merges(branch, old_revision, new_revision, merges, to_file)


def iter_parent_child(parent_map, revisions):
    for revision in revisions:
        parents = parent_map[revision]
        if len(parents) == 0:
            first_parent = _mod_revision.NULL_REVISION
        else:
            first_parent = parents[0]
        yield first_parent, revision


def iter_lca_revision(graph, revision, merge_revisions):
    for merge_revision in merge_revisions:
        base = graph.find_unique_lca(revision, merge_revision)
        yield base, merge_revision


def merges(repository, first_revision, second_revision):
    revision_ids = []
    for revision_id in repository.iter_reverse_revision_history(
        second_revision):
        if revision_id == first_revision:
            break
        revision_ids.append(revision_id)
    revision_ids.reverse()
    graph = repository.get_graph()
    parent_map = graph.get_parent_map(revision_ids)
    merge_revisions = [r for r in revision_ids if len(parent_map[r]) > 1]
    return list(iter_parent_child(parent_map, merge_revisions))
